// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

//
// This implementation supports persistence as well as the
// extended naming service with timestamps and dynamic updates.
//

package com.ooc.CosNaming;

import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;
import com.ooc.CosNaming.ExtNamingContextPackage.*;

//
// Not final because LocalNamingContext (ORBacus Names) inherits from it
//
public class NamingContext extends _ExtNamingContextImplBase
    implements Runnable
{
    static private int count_; // Running static count for keys

    private BOA boa_;
    private ORB orb_;
    private String key_;

    private java.util.Hashtable bindings_;
    private NamingContextSet ncs_;
    private NamingDatabase store_;
    boolean destroyed_; // Has the context been destroyed?

    //
    // Clients this service reports changes to
    //
    static private java.util.Vector listeners_ = new java.util.Vector();

    NamingContext(ORB orb, NamingDatabase store, NamingContextSet ncs,
		  String key)
    {
	orb_ = orb;
	store_ = store;
	ncs_ = ncs;
	key_ = key;
	bindings_ = new java.util.Hashtable(1023);

	((com.ooc.CORBA.ORB)orb_).connect(this, key_);

	if(ncs_ != null)
	    ncs_.add(this);

	if(store_ != null)
	{
	    store_.info();
	    store_.create(this, (int)(System.currentTimeMillis() / 1000));
	}
    }

    //
    // For ORBacus Names built-in naming service
    //
    public NamingContext(ORB orb, BOA boa)
    {
	this(orb, null, null, "DefaultNamingContext");

	boa_ = boa;
    }

    public void
    finalize()
    {
	if(ncs_ != null)
	    ncs_.remove(this);
    }

    // From java.lang.Thread, for ORBacus Names
    public void
    run()
    {
	boa_.impl_is_ready(null);
    }

    //
    // Add client to report changes to
    //
    public void
    add_client(com.ooc.CosNaming.BindingListener client)
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	java.util.Enumeration e = listeners_.elements();
	while(e.hasMoreElements())
	{
	    com.ooc.CosNaming.BindingListener l =
		(com.ooc.CosNaming.BindingListener)e.nextElement();

	    //
	    // Check if client is already registered
	    //
	    if(client._is_equivalent(l))
		return;
	}

	listeners_.addElement(client);
    }
 
    //
    // Remove client
    //
    public void
    remove_client(com.ooc.CosNaming.BindingListener client)
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	java.util.Enumeration e = listeners_.elements();
	while(e.hasMoreElements())
	{
	    com.ooc.CosNaming.BindingListener l =
		(com.ooc.CosNaming.BindingListener)e.nextElement();

	    if(client._is_equivalent(l))
	    {
		listeners_.removeElement(l);
		return;
	    }
	}
    }

    //
    // Bind without timestamp (standard mapping)
    //
    public void
    bind(NameComponent[] n, org.omg.CORBA.Object obj)
        throws NotFound, CannotProceed, InvalidName, AlreadyBound
    {
	bind(n, obj, (int)(System.currentTimeMillis() / 1000));
    }

    //
    // Bind with timestamp
    //
    public void
    bind(NameComponent[] n, org.omg.CORBA.Object obj, int timestamp)
        throws NotFound, CannotProceed, InvalidName, AlreadyBound
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	doBind(n, new NcOrObj(obj), false, timestamp);

        bindingAdded(n);
    }

    public void
    rebind(NameComponent[] n, org.omg.CORBA.Object obj)
        throws NotFound, CannotProceed, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	//
	// AlreadyBound must not be thrown here, so convert it.
	//
	try
	{
	    doBind(n, new NcOrObj(obj), true,
		   (int)(System.currentTimeMillis() / 1000));
	}
	catch(AlreadyBound ex)
	{
	    throw new CannotProceed(this, n);
	}

        bindingChanged(n, false);
    }

    //
    // Bind without timestamp (standard mapping)
    //
    public void
    bind_context(NameComponent[] n, org.omg.CosNaming.NamingContext nc)
        throws NotFound, CannotProceed, InvalidName, AlreadyBound
    {
	bind_context(n, nc, (int)(System.currentTimeMillis() / 1000));
    }

    //
    // Bind with timestamp
    //
    public void
    bind_context(NameComponent[] n, org.omg.CosNaming.NamingContext nc,
		 int timestamp)
        throws NotFound, CannotProceed, InvalidName, AlreadyBound
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	doBind(n, new NcOrObj(nc), false, timestamp);
 
        bindingAdded(n);
     }

    public void
    rebind_context(NameComponent[] n, org.omg.CosNaming.NamingContext nc)
        throws NotFound, CannotProceed, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	//
	// AlreadyBound must not be thrown here, so convert it.
	//
	try
	{
	    doBind(n, new NcOrObj(nc), true,
		   (int)(System.currentTimeMillis() / 1000));
	}
	catch(AlreadyBound ex)
	{
	    throw new CannotProceed(this, n);
	}

        bindingChanged(n, true);
    }

    public org.omg.CORBA.Object
    resolve(NameComponent[] n)
        throws NotFound, CannotProceed, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	int i;

	//
	// Check if name is valid
	//
	if(n.length == 0)
	    throw new InvalidName();
	if(n[0].id.length() == 0 && n[0].kind.length() == 0)
	    throw new InvalidName();
 
	NcOrObjBinding binding =
	    (NcOrObjBinding)bindings_.get(n[0].id + n[0].kind);

	//
	// Missing node if binding was not found
	//
	if(binding == null)
	    throw new NotFound(NotFoundReason.missing_node, n);

	if(n.length == 1)
	    return binding.ncOrObj.asObj();

	//
	// Binding must belong to a context
	//
	if(binding.ncOrObj.type() != BindingType.ncontext)
	    throw new NotFound(NotFoundReason.not_context, n);
 
	//
	// Context must not be a nil context here
	//
	if(binding.ncOrObj.nc() == null)
	    throw new CannotProceed(this, n);

	//
	// Create the rest of the name
	//
	NameComponent[] rest = new NameComponent[n.length - 1];
	System.arraycopy(n, 1, rest, 0, rest.length);

	//
	// Let the next context handle the resolving of the name
	//
	return binding.ncOrObj.nc().resolve(rest);
    }

    public void
    unbind(NameComponent[] n)
        throws NotFound, CannotProceed, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	int i;

	//
	// Check if name is valid
	//
	if(n.length == 0)
	    throw new InvalidName();
 	if(n[0].id.length() == 0 && n[0].kind.length() == 0)
	    throw new InvalidName();

	NcOrObjBinding binding =
	    (NcOrObjBinding)bindings_.get(n[0].id + n[0].kind);

	//
	// Missing node if binding was not found
	//
	if(binding == null)
	    throw new NotFound(NotFoundReason.missing_node, n);

	if(n.length == 1)
	{
	    //
	    // Do unbind
	    //
	    bindings_.remove(n[0].id + n[0].kind);
	    if(store_ != null)
		store_.unbind(this, n);

	    bindingRemoved(n, binding.ncOrObj.type() == BindingType.ncontext);
	}
	else
	{
	    //
	    // Binding must belong to a context
	    //
	    if(binding.ncOrObj.type() != BindingType.ncontext)
		throw new NotFound(NotFoundReason.not_context, n);
 
	    //
	    // Context must not be a nil context
	    //
	    if(binding.ncOrObj.nc() == null)
		throw new CannotProceed(this, n);

	    //
	    // Create the rest of the name
	    //
	    NameComponent[] rest = new NameComponent[n.length - 1];
	    System.arraycopy(n, 1, rest, 0, rest.length);

	    //
	    // Let the next context handle the resolving of the name
	    //
	    binding.ncOrObj.nc().unbind(rest);
	}
    }

    public org.omg.CosNaming.NamingContext
    new_context()
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	return new NamingContext(
	    orb_, store_, ncs_,
	    //
	    // Create new unique key. The key is composed of the current time
	    // and a running count.
	    //
	    new java.util.Date() + "-" + count_++);
    }

    public org.omg.CosNaming.NamingContext
    bind_new_context(NameComponent[] n)
        throws NotFound, AlreadyBound, CannotProceed, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	org.omg.CosNaming.NamingContext nc = new_context();
	bind_context(n, nc);

	return nc;
    }

    public void
    destroy()
        throws NotEmpty
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	if(!bindings_.isEmpty())
	    throw new NotEmpty();

	if(store_ != null)
	    store_.destroy(this);

	//
	// Remove ourself from the NamingContextSet. This will stop the
	// naming database from dumping this context.
	//
	ncs_.remove(this);
	destroyed_ = true;

	orb_.disconnect(this);
    }

    public void
    list(int how_many, BindingListHolder bl, BindingIteratorHolder bi)
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	int i;

	int len = bindings_.size();
	int num = len < how_many ? len : how_many;

	Binding[] b = new Binding[num];
	bl.value = b;

	java.util.Enumeration e = bindings_.elements();
	for(i = 0 ; i < num ; i++)
	{
	    NcOrObjBinding binding = (NcOrObjBinding)e.nextElement();
	    if(binding == null)
		throw new RuntimeException();

	    b[i] = new Binding();
	    b[i].binding_name = new NameComponent[1];
	    b[i].binding_name[0] = new NameComponent();
	    b[i].binding_name[0].id = binding.comp.id;
	    b[i].binding_name[0].kind = binding.comp.kind;
	    b[i].binding_type = binding.ncOrObj.type();
	}

	if(how_many < len)
	{
	    Binding[] rest = new Binding[len - how_many];

	    for(i = 0 ; i < len - how_many ; i++)
	    {
		NcOrObjBinding binding = (NcOrObjBinding)e.nextElement();
		if(binding == null)
		    throw new RuntimeException();

		rest[i] = new Binding();
		rest[i].binding_name = new NameComponent[1];
		rest[i].binding_name[0] = new NameComponent();
		rest[i].binding_name[0].id = binding.comp.id;
		rest[i].binding_name[0].kind = binding.comp.kind;
		rest[i].binding_type = binding.ncOrObj.type();
	    }

	    bi.value = new BindingIterator(orb_, rest);
	}
	else
	    bi.value = null;
    }

    String
    getKey()
    {
	return key_;
    }

    private void
    doBind(NameComponent[] n, NcOrObj ncOrObj, boolean rebind, int timestamp)
	throws NotFound, AlreadyBound, CannotProceed, InvalidName
    {
	//
	// Check if name is valid
	//
	if(n.length == 0)
	    throw new InvalidName();
	if(n[0].id.length() == 0 && n[0].kind.length() == 0)
	    throw new InvalidName();

	NcOrObjBinding binding =
	    (NcOrObjBinding)bindings_.get(n[0].id + n[0].kind);

	if(n.length == 1)
	{
	    //
	    // Name is already bound
	    //
	    if(!rebind && binding != null)
		throw new AlreadyBound();

	    if(store_ != null)
		store_.bind(this, n, ncOrObj.asObj(), ncOrObj.type(),
			    timestamp, rebind);

	    //
	    // Bind the name and the object
	    //
	    if(binding == null)
	    {
		binding = new NcOrObjBinding();
		bindings_.put(n[0].id + n[0].kind, binding);
	    }

	    binding.comp = n[0];
	    binding.ncOrObj = ncOrObj;
	    binding.timestamp = timestamp;
	}
	else
	{
	    //
	    // Missing node if binding was not found
	    //
	    if(binding == null)
		throw new NotFound(NotFoundReason.missing_node, n);

	    //
	    // Binding must belong to a context
	    //
	    if(binding.ncOrObj.type() != BindingType.ncontext)
		throw new NotFound(NotFoundReason.not_context, n);

	    //
	    // Context must not be a nil context here
	    //
	    if(binding.ncOrObj.nc() == null)
		throw new CannotProceed(this, n);

	    //
	    // Create the rest of the name
	    //
	    NameComponent[] rest = new NameComponent[n.length - 1];
	    System.arraycopy(n, 1, rest, 0, rest.length);

	    //
	    // Let the next context handle the binding
	    //
	    if(ncOrObj.type() == BindingType.ncontext)
	    {
		if(rebind)
		    binding.ncOrObj.nc().rebind_context(rest, ncOrObj.nc());
		else
		    binding.ncOrObj.nc().bind_context(rest, ncOrObj.nc());
	    }
	    else
	    {
		if(rebind)
		    binding.ncOrObj.nc().rebind(rest, ncOrObj.obj());
		else
		    binding.ncOrObj.nc().bind(rest, ncOrObj.obj());
	    }
	}
    }

    public void
    rename(NameComponent[] old_name, NameComponent[] new_name)
	throws NotFound, AlreadyBound, InvalidName
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	if(old_name.length != 1 || new_name.length != 1)
	    throw new InvalidName();

	NcOrObjBinding b =
	    (NcOrObjBinding)bindings_.get(old_name[0].id + old_name[0].kind);
	if(b == null)
	    throw new NotFound(NotFoundReason.missing_node, old_name);

	if(!b.comp.id.equals(new_name[0].id) ||
	   !b.comp.kind.equals(new_name[0].kind))
	{
	    //
	    // Check if new name already exists
	    //
	    if(bindings_.containsKey(new_name[0].id + new_name[0].kind))
		throw new AlreadyBound();

	    if(store_ != null)
		store_.unbind(this, old_name);

	    b.comp.id = new_name[0].id;
	    b.comp.kind = new_name[0].kind;
	    bindings_.remove(old_name[0].id + old_name[0].kind);
	    bindings_.put(new_name[0].id + new_name[0].kind, b);

	    if(store_ != null)
		store_.bind(this, new_name, b.ncOrObj.asObj(),
			    b.ncOrObj.type(),
			    (int)(System.currentTimeMillis() / 1000),
			    false);

	    bindingChanged(new_name, b.ncOrObj.type() == BindingType.ncontext);
	}
    }

    //
    // Get all extended bindings
    //
    public ExtendedBinding[]
    list_all()
    {
	if(destroyed_)
	    throw new OBJECT_NOT_EXIST();

	java.util.Enumeration e = bindings_.elements();
	int num = bindings_.size();

	ExtendedBinding[] eb = new ExtendedBinding[num];

	for(int i = 0 ; i < num ; i++)
	{
	    NcOrObjBinding binding = (NcOrObjBinding)e.nextElement();
	    if(binding == null)
		throw new RuntimeException();

	    NameComponent[] name = new NameComponent[1];
	    name[0] = binding.comp;
	    eb[i] = new ExtendedBinding(name, binding.ncOrObj.type(),
					binding.timestamp);
	}

	return eb;
    }

    //
    // Binding was added
    //
    private void
    bindingAdded(NameComponent[] n)
    {
	//
	// Check whether change occured in this context
	//
	if(n.length != 1)
	    return;

	NcOrObjBinding binding =
	    (NcOrObjBinding)bindings_.get(n[0].id + n[0].kind);
	boolean isContext = (binding.ncOrObj.type() == BindingType.ncontext);

	java.util.Enumeration e = listeners_.elements();
	while(e.hasMoreElements())
	{
	    com.ooc.CosNaming.BindingListener l =
		(com.ooc.CosNaming.BindingListener)e.nextElement();

	    try
	    {
		if(isContext)
		    l.context_added(this, n[0]);
		else
		    l.object_added(this, n[0]);
	    }
	    catch(COMM_FAILURE ex)
	    {
		remove_client(l);
	    }
	}
    }

    //
    // Binding was removed
    //
    private void
    bindingRemoved(NameComponent[] n, boolean isContext)
    {
	//
	// Check whether change occured in this context
	//
	if(n.length != 1)
	    return;

	java.util.Enumeration e = listeners_.elements();
	while(e.hasMoreElements())
	{
	    com.ooc.CosNaming.BindingListener l =
		(com.ooc.CosNaming.BindingListener)e.nextElement();

	    try
	    {
		if(isContext)
		    l.context_removed(this, n[0]);
		else
		    l.object_removed(this, n[0]);
	    }
	    catch(COMM_FAILURE ex)
	    {
		remove_client(l);
	    }
	}
    }

    //
    // Binding was changed
    //
    private void
    bindingChanged(NameComponent[] name, boolean isContext)
    {
	//
	// Check whether change occured in this context
	//
	if(name.length != 1)
	    return;

	java.util.Enumeration e = listeners_.elements();
	while(e.hasMoreElements())
	{
	    com.ooc.CosNaming.BindingListener l =
		(com.ooc.CosNaming.BindingListener)e.nextElement();

	    try
	    {
		if(isContext)
		    l.context_changed(this, name[0]);
		else
		    l.object_changed(this, name[0]);
	    }
	    catch(COMM_FAILURE ex)
	    {
		remove_client(l);
	    }
	}
    }


    //
    // NcOrObj and NcOrObjBinding
    //

    class NcOrObj
    {
	private org.omg.CORBA.Object obj_;
	private org.omg.CosNaming.NamingContext context_;
	private BindingType type_;

	NcOrObj(org.omg.CORBA.Object obj)
	{
	    obj_ = obj;
	    type_ = BindingType.nobject;
	}

	NcOrObj(org.omg.CosNaming.NamingContext context)
	{
	    context_ = context;
	    type_ = BindingType.ncontext;
	}

	org.omg.CORBA.Object
	obj()
	{
	    if(type_ != BindingType.nobject)
		throw new RuntimeException();

	    return obj_;
	}

	org.omg.CosNaming.NamingContext
	nc()
	{
	    if(type_ != BindingType.ncontext)
		throw new RuntimeException();

	    return context_;
	}

	BindingType
	type()
	{
	    return type_;
	}

	org.omg.CORBA.Object
	asObj()
	{
	    if(type_ == BindingType.nobject)
		return obj_;
	    else
		return context_;
	}
    }

    class NcOrObjBinding
    {
	private NcOrObj ncOrObj;
	private org.omg.CosNaming.NameComponent comp;
	private int timestamp;
    }
}
